# This file was auto-generated by Fern from our API Definition.

import typing

from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from ..types.json_node import JsonNode
from ..types.local_runner import LocalRunner
from ..types.local_runner_connect_response import LocalRunnerConnectResponse
from ..types.local_runner_heartbeat_response import LocalRunnerHeartbeatResponse
from ..types.local_runner_job import LocalRunnerJob
from ..types.local_runner_job_metadata import LocalRunnerJobMetadata
from ..types.local_runner_job_page import LocalRunnerJobPage
from ..types.local_runner_log_entry import LocalRunnerLogEntry
from ..types.local_runner_page import LocalRunnerPage
from ..types.local_runner_pair_response import LocalRunnerPairResponse
from .raw_client import AsyncRawRunnersClient, RawRunnersClient
from .types.local_runner_job_result_request_status import LocalRunnerJobResultRequestStatus

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RunnersClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawRunnersClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawRunnersClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawRunnersClient
        """
        return self._raw_client

    def get_job_logs(
        self,
        job_id: str,
        *,
        offset: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> typing.List[LocalRunnerLogEntry]:
        """
        Get log entries for a local runner job

        Parameters
        ----------
        job_id : str

        offset : typing.Optional[int]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.List[LocalRunnerLogEntry]
            Log entries

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.runners.get_job_logs(job_id='jobId', )
        """
        _response = self._raw_client.get_job_logs(job_id, offset=offset, request_options=request_options)
        return _response.data

    def append_job_logs(
        self,
        job_id: str,
        *,
        request: typing.Sequence[LocalRunnerLogEntry],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Append log entries for a running local runner job

        Parameters
        ----------
        job_id : str

        request : typing.Sequence[LocalRunnerLogEntry]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        from Opik import LocalRunnerLogEntry
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.runners.append_job_logs(job_id='jobId', request=[LocalRunnerLogEntry(stream='stream', text='text', )], )
        """
        _response = self._raw_client.append_job_logs(job_id, request=request, request_options=request_options)
        return _response.data

    def cancel_job(self, job_id: str, *, request_options: typing.Optional[RequestOptions] = None) -> None:
        """
        Cancel a pending or running local runner job

        Parameters
        ----------
        job_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.runners.cancel_job(job_id='jobId', )
        """
        _response = self._raw_client.cancel_job(job_id, request_options=request_options)
        return _response.data

    def connect_runner(
        self, *, pairing_code: str, runner_name: str, request_options: typing.Optional[RequestOptions] = None
    ) -> LocalRunnerConnectResponse:
        """
        Exchange a pairing code or API key for local runner credentials

        Parameters
        ----------
        pairing_code : str

        runner_name : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        LocalRunnerConnectResponse
            Runner connected

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.runners.connect_runner(pairing_code='pairing_code', runner_name='runner_name', )
        """
        _response = self._raw_client.connect_runner(
            pairing_code=pairing_code, runner_name=runner_name, request_options=request_options
        )
        return _response.data

    def create_job(
        self,
        *,
        agent_name: str,
        project_id: str,
        inputs: typing.Optional[JsonNode] = OMIT,
        mask_id: typing.Optional[str] = OMIT,
        metadata: typing.Optional[LocalRunnerJobMetadata] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Create a local runner job and enqueue it for execution

        Parameters
        ----------
        agent_name : str

        project_id : str

        inputs : typing.Optional[JsonNode]

        mask_id : typing.Optional[str]

        metadata : typing.Optional[LocalRunnerJobMetadata]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.runners.create_job(agent_name='agent_name', project_id='project_id', )
        """
        _response = self._raw_client.create_job(
            agent_name=agent_name,
            project_id=project_id,
            inputs=inputs,
            mask_id=mask_id,
            metadata=metadata,
            request_options=request_options,
        )
        return _response.data

    def generate_pairing_code(
        self, *, project_id: str, request_options: typing.Optional[RequestOptions] = None
    ) -> LocalRunnerPairResponse:
        """
        Generate a pairing code for a local runner in the current workspace

        Parameters
        ----------
        project_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        LocalRunnerPairResponse
            Pairing code generated

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.runners.generate_pairing_code(project_id='project_id', )
        """
        _response = self._raw_client.generate_pairing_code(project_id=project_id, request_options=request_options)
        return _response.data

    def get_job(self, job_id: str, *, request_options: typing.Optional[RequestOptions] = None) -> LocalRunnerJob:
        """
        Get a single local runner job's status and results

        Parameters
        ----------
        job_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        LocalRunnerJob
            Job details

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.runners.get_job(job_id='jobId', )
        """
        _response = self._raw_client.get_job(job_id, request_options=request_options)
        return _response.data

    def get_runner(self, runner_id: str, *, request_options: typing.Optional[RequestOptions] = None) -> LocalRunner:
        """
        Get a single local runner with its registered agents

        Parameters
        ----------
        runner_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        LocalRunner
            Runner details

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.runners.get_runner(runner_id='runnerId', )
        """
        _response = self._raw_client.get_runner(runner_id, request_options=request_options)
        return _response.data

    def heartbeat(
        self, runner_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> LocalRunnerHeartbeatResponse:
        """
        Refresh local runner heartbeat

        Parameters
        ----------
        runner_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        LocalRunnerHeartbeatResponse
            Heartbeat response

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.runners.heartbeat(runner_id='runnerId', )
        """
        _response = self._raw_client.heartbeat(runner_id, request_options=request_options)
        return _response.data

    def list_jobs(
        self,
        runner_id: str,
        *,
        project_id: typing.Optional[str] = None,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> LocalRunnerJobPage:
        """
        List jobs for a local runner

        Parameters
        ----------
        runner_id : str

        project_id : typing.Optional[str]

        page : typing.Optional[int]

        size : typing.Optional[int]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        LocalRunnerJobPage
            Jobs list

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.runners.list_jobs(runner_id='runnerId', )
        """
        _response = self._raw_client.list_jobs(
            runner_id, project_id=project_id, page=page, size=size, request_options=request_options
        )
        return _response.data

    def list_runners(
        self,
        *,
        project_id: str,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> LocalRunnerPage:
        """
        List local runners owned by the current user in the workspace

        Parameters
        ----------
        project_id : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        LocalRunnerPage
            Runners list

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.runners.list_runners(project_id='project_id', )
        """
        _response = self._raw_client.list_runners(
            project_id=project_id, page=page, size=size, request_options=request_options
        )
        return _response.data

    def next_job(self, runner_id: str, *, request_options: typing.Optional[RequestOptions] = None) -> LocalRunnerJob:
        """
        Long-poll for the next pending local runner job

        Parameters
        ----------
        runner_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        LocalRunnerJob
            Job available

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.runners.next_job(runner_id='runnerId', )
        """
        _response = self._raw_client.next_job(runner_id, request_options=request_options)
        return _response.data

    def register_agents(
        self,
        runner_id: str,
        *,
        request: typing.Dict[str, typing.Optional[typing.Any]],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Register or update the local runner's agent list

        Parameters
        ----------
        runner_id : str

        request : typing.Dict[str, typing.Optional[typing.Any]]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.runners.register_agents(runner_id='runnerId', request={'key': 'value'
        }, )
        """
        _response = self._raw_client.register_agents(runner_id, request=request, request_options=request_options)
        return _response.data

    def report_job_result(
        self,
        job_id: str,
        *,
        status: LocalRunnerJobResultRequestStatus,
        result: typing.Optional[JsonNode] = OMIT,
        error: typing.Optional[str] = OMIT,
        trace_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Report local runner job completion or failure

        Parameters
        ----------
        job_id : str

        status : LocalRunnerJobResultRequestStatus

        result : typing.Optional[JsonNode]

        error : typing.Optional[str]

        trace_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.runners.report_job_result(job_id='jobId', status="pending", )
        """
        _response = self._raw_client.report_job_result(
            job_id, status=status, result=result, error=error, trace_id=trace_id, request_options=request_options
        )
        return _response.data


class AsyncRunnersClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawRunnersClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawRunnersClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawRunnersClient
        """
        return self._raw_client

    async def get_job_logs(
        self,
        job_id: str,
        *,
        offset: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> typing.List[LocalRunnerLogEntry]:
        """
        Get log entries for a local runner job

        Parameters
        ----------
        job_id : str

        offset : typing.Optional[int]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.List[LocalRunnerLogEntry]
            Log entries

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.runners.get_job_logs(job_id='jobId', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_job_logs(job_id, offset=offset, request_options=request_options)
        return _response.data

    async def append_job_logs(
        self,
        job_id: str,
        *,
        request: typing.Sequence[LocalRunnerLogEntry],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Append log entries for a running local runner job

        Parameters
        ----------
        job_id : str

        request : typing.Sequence[LocalRunnerLogEntry]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        from Opik import LocalRunnerLogEntry
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.runners.append_job_logs(job_id='jobId', request=[LocalRunnerLogEntry(stream='stream', text='text', )], )
        asyncio.run(main())
        """
        _response = await self._raw_client.append_job_logs(job_id, request=request, request_options=request_options)
        return _response.data

    async def cancel_job(self, job_id: str, *, request_options: typing.Optional[RequestOptions] = None) -> None:
        """
        Cancel a pending or running local runner job

        Parameters
        ----------
        job_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.runners.cancel_job(job_id='jobId', )
        asyncio.run(main())
        """
        _response = await self._raw_client.cancel_job(job_id, request_options=request_options)
        return _response.data

    async def connect_runner(
        self, *, pairing_code: str, runner_name: str, request_options: typing.Optional[RequestOptions] = None
    ) -> LocalRunnerConnectResponse:
        """
        Exchange a pairing code or API key for local runner credentials

        Parameters
        ----------
        pairing_code : str

        runner_name : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        LocalRunnerConnectResponse
            Runner connected

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.runners.connect_runner(pairing_code='pairing_code', runner_name='runner_name', )
        asyncio.run(main())
        """
        _response = await self._raw_client.connect_runner(
            pairing_code=pairing_code, runner_name=runner_name, request_options=request_options
        )
        return _response.data

    async def create_job(
        self,
        *,
        agent_name: str,
        project_id: str,
        inputs: typing.Optional[JsonNode] = OMIT,
        mask_id: typing.Optional[str] = OMIT,
        metadata: typing.Optional[LocalRunnerJobMetadata] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Create a local runner job and enqueue it for execution

        Parameters
        ----------
        agent_name : str

        project_id : str

        inputs : typing.Optional[JsonNode]

        mask_id : typing.Optional[str]

        metadata : typing.Optional[LocalRunnerJobMetadata]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.runners.create_job(agent_name='agent_name', project_id='project_id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.create_job(
            agent_name=agent_name,
            project_id=project_id,
            inputs=inputs,
            mask_id=mask_id,
            metadata=metadata,
            request_options=request_options,
        )
        return _response.data

    async def generate_pairing_code(
        self, *, project_id: str, request_options: typing.Optional[RequestOptions] = None
    ) -> LocalRunnerPairResponse:
        """
        Generate a pairing code for a local runner in the current workspace

        Parameters
        ----------
        project_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        LocalRunnerPairResponse
            Pairing code generated

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.runners.generate_pairing_code(project_id='project_id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.generate_pairing_code(project_id=project_id, request_options=request_options)
        return _response.data

    async def get_job(self, job_id: str, *, request_options: typing.Optional[RequestOptions] = None) -> LocalRunnerJob:
        """
        Get a single local runner job's status and results

        Parameters
        ----------
        job_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        LocalRunnerJob
            Job details

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.runners.get_job(job_id='jobId', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_job(job_id, request_options=request_options)
        return _response.data

    async def get_runner(
        self, runner_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> LocalRunner:
        """
        Get a single local runner with its registered agents

        Parameters
        ----------
        runner_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        LocalRunner
            Runner details

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.runners.get_runner(runner_id='runnerId', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_runner(runner_id, request_options=request_options)
        return _response.data

    async def heartbeat(
        self, runner_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> LocalRunnerHeartbeatResponse:
        """
        Refresh local runner heartbeat

        Parameters
        ----------
        runner_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        LocalRunnerHeartbeatResponse
            Heartbeat response

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.runners.heartbeat(runner_id='runnerId', )
        asyncio.run(main())
        """
        _response = await self._raw_client.heartbeat(runner_id, request_options=request_options)
        return _response.data

    async def list_jobs(
        self,
        runner_id: str,
        *,
        project_id: typing.Optional[str] = None,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> LocalRunnerJobPage:
        """
        List jobs for a local runner

        Parameters
        ----------
        runner_id : str

        project_id : typing.Optional[str]

        page : typing.Optional[int]

        size : typing.Optional[int]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        LocalRunnerJobPage
            Jobs list

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.runners.list_jobs(runner_id='runnerId', )
        asyncio.run(main())
        """
        _response = await self._raw_client.list_jobs(
            runner_id, project_id=project_id, page=page, size=size, request_options=request_options
        )
        return _response.data

    async def list_runners(
        self,
        *,
        project_id: str,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> LocalRunnerPage:
        """
        List local runners owned by the current user in the workspace

        Parameters
        ----------
        project_id : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        LocalRunnerPage
            Runners list

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.runners.list_runners(project_id='project_id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.list_runners(
            project_id=project_id, page=page, size=size, request_options=request_options
        )
        return _response.data

    async def next_job(
        self, runner_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> LocalRunnerJob:
        """
        Long-poll for the next pending local runner job

        Parameters
        ----------
        runner_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        LocalRunnerJob
            Job available

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.runners.next_job(runner_id='runnerId', )
        asyncio.run(main())
        """
        _response = await self._raw_client.next_job(runner_id, request_options=request_options)
        return _response.data

    async def register_agents(
        self,
        runner_id: str,
        *,
        request: typing.Dict[str, typing.Optional[typing.Any]],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Register or update the local runner's agent list

        Parameters
        ----------
        runner_id : str

        request : typing.Dict[str, typing.Optional[typing.Any]]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.runners.register_agents(runner_id='runnerId', request={'key': 'value'
            }, )
        asyncio.run(main())
        """
        _response = await self._raw_client.register_agents(runner_id, request=request, request_options=request_options)
        return _response.data

    async def report_job_result(
        self,
        job_id: str,
        *,
        status: LocalRunnerJobResultRequestStatus,
        result: typing.Optional[JsonNode] = OMIT,
        error: typing.Optional[str] = OMIT,
        trace_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Report local runner job completion or failure

        Parameters
        ----------
        job_id : str

        status : LocalRunnerJobResultRequestStatus

        result : typing.Optional[JsonNode]

        error : typing.Optional[str]

        trace_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.runners.report_job_result(job_id='jobId', status="pending", )
        asyncio.run(main())
        """
        _response = await self._raw_client.report_job_result(
            job_id, status=status, result=result, error=error, trace_id=trace_id, request_options=request_options
        )
        return _response.data



import os

from setuptools import find_packages, setup

project_urls = {"Source code": "https://github.com/comet-ml/opik"}

HERE = os.path.abspath(os.path.dirname(__file__))
version = os.environ.get("VERSION")
if version is None:
    version_file = os.path.join(HERE, "..", "..", "version.txt")
    if os.path.exists(version_file):
        with open(version_file) as fp:
            version = fp.read().strip()
    else:
        version = "0.0.1"

setup(
    author="Comet ML Inc.",
    author_email="mail@comet.com",
    python_requires=">=3.10",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
    ],
    description="Comet tool for logging and evaluating LLM traces",
    long_description=open(
        os.path.join(HERE, "..", "..", "README.md"), encoding="utf-8"
    ).read(),
    long_description_content_type="text/markdown",
    install_requires=[
        "boto3-stubs[bedrock-runtime]>=1.34.110",
        "click",
        "httpx",  # some older version of openai/litellm are broken with httpx>=0.28.0
        "rapidfuzz>=3.0.0,<4.0.0",
        # LiteLLM dependency comments:
        # - Exclude litellm 1.75.0-1.75.5 (broken callbacks system)
        # - Exclude versions 1.77.2-1.77.4: introduce C++ compiler dependency (madoka), fixed in 1.77.5
        #   See: https://github.com/BerriAI/litellm/issues/14762
        # - Exclude versions 1.77.5-1.79.1: remove trace_id/parent_span_id passthrough, fixed in 1.79.2+
        #   See: https://github.com/BerriAI/litellm/pull/15529
        # Please keep this list in sync with the one in sdks/opik_optimizer/pyproject.toml
        "litellm>=1.79.2,!=1.75.0,!=1.75.1,!=1.75.2,!=1.75.3,!=1.75.4,!=1.75.5,!=1.77.3,!=1.77.4,!=1.77.5,!=1.77.7,!=1.78.0,!=1.78.2,!=1.78.3,!=1.78.4,!=1.78.5,!=1.78.6,!=1.78.7,!=1.79.0,!=1.79.1",
        "openai",
        "pydantic-settings>=2.0.0,<3.0.0,!=2.9.0",
        "pydantic>=2.0.0,<3.0.0",
        "pytest",
        "rich",
        "sentry_sdk>=2.0.0",
        "tenacity",
        "tqdm",
        "uuid6",
        "jinja2",
    ],
    extras_require={
        "proxy": [
            "fastapi>=0.100.0",
            "uvicorn>=0.23.0",
        ],
    },
    entry_points={
        "pytest11": [
            "opik = opik.plugins.pytest.hooks",
        ],
        "console_scripts": ["opik = opik.cli:cli"],
    },
    keywords="opik",
    name="opik",
    include_package_data=True,
    package_data={"opik": ["py.typed"]},
    packages=find_packages("src"),
    package_dir={"": "src"},
    url="https://www.comet.com",
    project_urls=project_urls,
    version=version,
    zip_safe=False,
    license="Apache 2.0 License",
)



import logging
from typing import Any, Dict, List, Optional

import litellm.types.utils

LOGGER = logging.getLogger(__name__)


def _initialize_aggregated_response(
    first_chunk: litellm.types.utils.ModelResponse,
) -> Dict[str, Any]:
    return {
        "choices": [{"index": 0, "message": {"role": "", "content": ""}}],
        "created": getattr(first_chunk, "created", 0),
        "id": getattr(first_chunk, "id", ""),
        "model": getattr(first_chunk, "model", ""),
        "object": "chat.completion",
        "system_fingerprint": getattr(first_chunk, "system_fingerprint", None),
        "usage": None,
    }


def _extract_role_from_delta(delta: Any, current_role: str) -> str:
    if hasattr(delta, "role") and delta.role and not current_role:
        return delta.role
    return current_role


def _extract_content_from_delta(delta: Any) -> Optional[str]:
    if hasattr(delta, "content") and delta.content:
        return delta.content
    return None


def _extract_finish_reason_from_choice(choice: Any) -> Optional[str]:
    if hasattr(choice, "finish_reason") and choice.finish_reason:
        return choice.finish_reason
    return None


def _extract_usage_from_chunk(
    chunk: litellm.types.utils.ModelResponse,
) -> Optional[Dict[str, Any]]:
    if not hasattr(chunk, "usage") or chunk.usage is None:
        return None

    try:
        if hasattr(chunk.usage, "model_dump"):
            usage_dict = chunk.usage.model_dump()
        elif hasattr(chunk.usage, "dict"):
            usage_dict = chunk.usage.dict()
        elif isinstance(chunk.usage, dict):
            usage_dict = chunk.usage
        else:
            return None

        if usage_dict and isinstance(usage_dict, dict):
            filtered_usage = {k: v for k, v in usage_dict.items() if v is not None}
            return filtered_usage if filtered_usage else None
        return None
    except Exception as exception:
        LOGGER.debug(
            "Error extracting usage from streaming chunk: %s",
            str(exception),
            exc_info=True,
        )
        return None


def aggregate(
    items: List[litellm.types.utils.ModelResponse],
) -> Optional[litellm.types.utils.ModelResponse]:
    try:
        if not items:
            return None

        aggregated_response = _initialize_aggregated_response(items[0])
        text_chunks: List[str] = []

        for chunk in items:
            if not hasattr(chunk, "choices") or not chunk.choices:
                continue

            choice = chunk.choices[0]

            if hasattr(choice, "delta") and choice.delta:
                delta = choice.delta

                current_role = aggregated_response["choices"][0]["message"]["role"]
                aggregated_response["choices"][0]["message"]["role"] = (
                    _extract_role_from_delta(delta, current_role)
                )

                content = _extract_content_from_delta(delta)
                if content:
                    text_chunks.append(content)

            finish_reason = _extract_finish_reason_from_choice(choice)
            if finish_reason:
                aggregated_response["choices"][0]["finish_reason"] = finish_reason

            chunk_usage = _extract_usage_from_chunk(chunk)
            if chunk_usage:
                aggregated_response["usage"] = chunk_usage

        aggregated_response["choices"][0]["message"]["content"] = "".join(text_chunks)
        return litellm.types.utils.ModelResponse(**aggregated_response)

    except Exception as exception:
        LOGGER.error(
            "Failed to aggregate LiteLLM streaming chunks: %s",
            str(exception),
            exc_info=True,
        )
        return None



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .trace_thread_filter_public import TraceThreadFilterPublic
from .trace_thread_user_defined_metric_python_code_public import TraceThreadUserDefinedMetricPythonCodePublic


class AutomationRuleEvaluatorTraceThreadUserDefinedMetricPythonPublic(UniversalBaseModel):
    filters: typing.Optional[typing.List[TraceThreadFilterPublic]] = None
    code: typing.Optional[TraceThreadUserDefinedMetricPythonCodePublic] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing

LocalRunnerJobStatus = typing.Union[
    typing.Literal["pending", "running", "completed", "failed", "cancelled"], typing.Any
]



import anthropic
import logging
import functools
from typing import (
    Optional,
    Callable,
    Iterator,
    AsyncIterator,
    Any,
)
from opik.types import ErrorInfoDict
from opik.api_objects import trace, span
from opik.decorator import generator_wrappers, error_info_collector

from anthropic.lib.streaming import _messages

LOGGER = logging.getLogger(__name__)

original_stream_iter_method = anthropic.Stream.__iter__
original_async_stream_aiter_method = anthropic.AsyncStream.__aiter__

original_message_stream_iter_method = anthropic.MessageStream.__iter__
original_async_message_stream_aiter_method = anthropic.AsyncMessageStream.__aiter__

original_message_stream_manager_enter_method = anthropic.MessageStreamManager.__enter__
original_async_message_stream_manager_aenter_method = (
    anthropic.AsyncMessageStreamManager.__aenter__
)

try:
    from anthropic.lib.streaming._beta_messages import (
        BetaAsyncMessageStream,
        BetaAsyncMessageStreamManager,
        BetaMessageStream,
        BetaMessageStreamManager,
    )

    original_beta_message_stream_iter_method = BetaMessageStream.__iter__
    original_beta_async_message_stream_aiter_method = BetaAsyncMessageStream.__aiter__
    original_beta_message_stream_manager_enter_method = (
        BetaMessageStreamManager.__enter__
    )
    original_beta_async_message_stream_manager_aenter_method = (
        BetaAsyncMessageStreamManager.__aenter__
    )
except ImportError:
    LOGGER.debug(
        "Anthropic beta streaming module not available; beta stream patching disabled."
    )
    BetaMessageStream = None  # type: ignore[assignment,misc]
    BetaAsyncMessageStream = None  # type: ignore[assignment,misc]
    BetaMessageStreamManager = None  # type: ignore[assignment,misc]
    BetaAsyncMessageStreamManager = None  # type: ignore[assignment,misc]
    original_beta_message_stream_iter_method = None
    original_beta_async_message_stream_aiter_method = None
    original_beta_message_stream_manager_enter_method = None
    original_beta_async_message_stream_manager_aenter_method = None


def patch_sync_stream(
    stream: anthropic.Stream,
    span_to_end: span.SpanData,
    trace_to_end: Optional[trace.TraceData],
    finally_callback: generator_wrappers.FinishGeneratorCallback,
) -> anthropic.Stream:
    """
    Used in the following cases
    ```
    stream = client.messages.create(stream=True)
    for event in stream:
        print(event)
    ```
    """

    def Stream__iter__decorator(dunder_iter_func: Callable) -> Callable:
        @functools.wraps(dunder_iter_func)
        def wrapper(
            self: anthropic.Stream,
        ) -> Iterator[Any]:
            try:
                accumulated_message = None
                error_info: Optional[ErrorInfoDict] = None

                for item in dunder_iter_func(self):
                    accumulated_message = _messages.accumulate_event(
                        event=item, current_snapshot=accumulated_message
                    )
                    yield item
            except Exception as exception:
                LOGGER.debug(
                    "Exception raised from anthropic.Stream.",
                    str(exception),
                    exc_info=True,
                )
                error_info = error_info_collector.collect(exception)
                raise exception
            finally:
                if not hasattr(self, "opik_tracked_instance"):
                    return

                delattr(self, "opik_tracked_instance")
                output = accumulated_message if error_info is None else None
                finally_callback(
                    output=output,
                    error_info=error_info,
                    capture_output=True,
                    generators_span_to_end=self.span_to_end,
                    generators_trace_to_end=self.trace_to_end,
                )

        return wrapper

    anthropic.Stream.__iter__ = Stream__iter__decorator(original_stream_iter_method)

    stream.opik_tracked_instance = True
    stream.span_to_end = span_to_end
    stream.trace_to_end = trace_to_end

    return stream


def patch_async_stream(
    stream: anthropic.AsyncStream,
    span_to_end: span.SpanData,
    trace_to_end: Optional[trace.TraceData],
    finally_callback: generator_wrappers.FinishGeneratorCallback,
) -> anthropic.Stream:
    """
    Used in the following cases
    ```
    astream = async_client.messages.create(stream=True)
    async for event in astream:
        print(event)
    ```
    """

    def AsyncStream__aiter__decorator(dunder_aiter_func: Callable) -> Callable:
        @functools.wraps(dunder_aiter_func)
        async def wrapper(
            self: anthropic.AsyncStream,
        ) -> AsyncIterator[Any]:
            try:
                accumulated_message = None
                error_info: Optional[ErrorInfoDict] = None

                async for item in dunder_aiter_func(self):
                    accumulated_message = _messages.accumulate_event(
                        event=item, current_snapshot=accumulated_message
                    )
                    yield item
            except Exception as exception:
                LOGGER.debug(
                    "Exception raised from anthropic.AsyncStream.",
                    str(exception),
                    exc_info=True,
                )
                error_info = error_info_collector.collect(exception)
                raise exception
            finally:
                if not hasattr(self, "opik_tracked_instance"):
                    return

                delattr(self, "opik_tracked_instance")
                output = accumulated_message if error_info is None else None
                finally_callback(
                    output=output,
                    error_info=error_info,
                    capture_output=True,
                    generators_span_to_end=self.span_to_end,
                    generators_trace_to_end=self.trace_to_end,
                )

        return wrapper

    anthropic.AsyncStream.__aiter__ = AsyncStream__aiter__decorator(
        original_async_stream_aiter_method
    )

    stream.opik_tracked_instance = True
    stream.span_to_end = span_to_end
    stream.trace_to_end = trace_to_end

    return stream


def patch_sync_message_stream_manager(
    message_stream_manager: anthropic.MessageStreamManager,
    span_to_end: span.SpanData,
    trace_to_end: Optional[trace.TraceData],
    finally_callback: generator_wrappers.FinishGeneratorCallback,
) -> anthropic.MessageStreamManager:
    """
    User flow that caused this non-trivial patching:

    ```
    stream_manager = anthropic_client.messages.stream(...)

    with stream_manager as stream:
        for event in stream:
            print(event)

        stream.get_final_message()
    ```

    `stream` method returns an object (stream_manager), that creates a MessageStream object in the context
    manager. MessageStream class has it's own public API (e.g. get_final_message), so we can't replace it with our generator.
    We need to patch __iter__ method of MessageStream, so that when it returns a generator object, we
    wrap it.

    In addition, its possible that generator is used multiple times. We are making sure, that we execute our
    logging logic only once.
    """

    def MessageStream__iter__decorator(dunder_iter_func: Callable) -> Callable:
        @functools.wraps(dunder_iter_func)
        def wrapper(
            self: anthropic.MessageStream,
        ) -> Iterator[anthropic.MessageStreamEvent]:
            try:
                error_info: Optional[ErrorInfoDict] = None
                for item in dunder_iter_func(self):
                    yield item
            except Exception as exception:
                LOGGER.debug(
                    "Exception raised from anthropic.MessageStream.",
                    str(exception),
                    exc_info=True,
                )
                error_info = error_info_collector.collect(exception)
                raise exception
            finally:
                if not hasattr(self, "opik_tracked_instance"):
                    return

                delattr(self, "opik_tracked_instance")

                accumulated_output = (
                    self.get_final_message() if error_info is None else None
                )

                finally_callback(
                    output=accumulated_output,
                    error_info=error_info,
                    capture_output=True,
                    generators_span_to_end=self.span_to_end,
                    generators_trace_to_end=self.trace_to_end,
                )

        return wrapper

    def MessageStreamManager__enter__decorator(dunder_enter_func: Callable) -> Callable:
        @functools.wraps(dunder_enter_func)
        def wrapper(self: anthropic.MessageStreamManager) -> anthropic.MessageStream:
            result: anthropic.MessageStream = dunder_enter_func(self)

            if hasattr(self, "opik_tracked_instance"):
                result.opik_tracked_instance = True
                result.span_to_end = self.span_to_end
                result.trace_to_end = self.trace_to_end

            return result

        return wrapper

    # We are decorating class methods instead of instance methods because
    # python interpreter often (if not always) looks for dunder methods for in classes, not instances, by .
    # Decorating an instance method will not work, original method will always be called.
    anthropic.MessageStreamManager.__enter__ = MessageStreamManager__enter__decorator(
        original_message_stream_manager_enter_method
    )
    anthropic.MessageStream.__iter__ = MessageStream__iter__decorator(
        original_message_stream_iter_method
    )

    message_stream_manager.opik_tracked_instance = True
    message_stream_manager.span_to_end = span_to_end
    message_stream_manager.trace_to_end = trace_to_end

    return message_stream_manager


def patch_async_message_stream_manager(
    async_message_stream_manager: anthropic.AsyncMessageStreamManager,
    span_to_end: span.SpanData,
    trace_to_end: Optional[trace.TraceData],
    finally_callback: generator_wrappers.FinishGeneratorCallback,
) -> anthropic.AsyncMessageStreamManager:
    """
    User flow that caused this non-trivial patching:

    ```
    async_stream_manager = async_anthropic_client.messages.stream(...)

    async with async_stream_manager as async_stream:
        async for event in async_stream:
            print(event)

        await stream.get_final_message()
    ```

    For more details see patch_sync_message_stream_manager docstring
    """

    def AsyncMessageStream__aiter__decorator(dunder_aiter_func: Callable) -> Callable:
        @functools.wraps(dunder_aiter_func)
        async def wrapper(
            self: anthropic.AsyncMessageStream,
        ) -> AsyncIterator[anthropic.MessageStreamEvent]:
            try:
                error_info: Optional[ErrorInfoDict] = None
                async for item in dunder_aiter_func(self):
                    yield item
            except Exception as exception:
                LOGGER.debug(
                    "Exception raised from anthropic.AsyncMessageStream.",
                    str(exception),
                    exc_info=True,
                )
                error_info = error_info_collector.collect(exception)
                raise exception
            finally:
                if not hasattr(self, "opik_tracked_instance"):
                    return

                delattr(self, "opik_tracked_instance")

                accumulated_output = (
                    await self.get_final_message() if error_info is None else None
                )

                finally_callback(
                    output=accumulated_output,
                    error_info=error_info,
                    capture_output=True,
                    generators_span_to_end=self.span_to_end,
                    generators_trace_to_end=self.trace_to_end,
                )

        return wrapper

    def AsyncMessageStreamManager__aenter__decorator(
        dunder_aenter_func: Callable,
    ) -> Callable:
        @functools.wraps(dunder_aenter_func)
        async def wrapper(
            self: anthropic.AsyncMessageStreamManager,
        ) -> anthropic.AsyncMessageStream:
            result: anthropic.AsyncMessageStream = await dunder_aenter_func(self)

            if hasattr(self, "opik_tracked_instance"):
                result.opik_tracked_instance = True
                result.span_to_end = self.span_to_end
                result.trace_to_end = self.trace_to_end

            return result

        return wrapper

    # We are decorating class methods instead of instance methods because
    # python interpreter often (if not always) looks for dunder methods for in classes, not instances, by .
    # Decorating an instance method will not work, original method will always be called.
    anthropic.AsyncMessageStreamManager.__aenter__ = (
        AsyncMessageStreamManager__aenter__decorator(
            original_async_message_stream_manager_aenter_method
        )
    )
    anthropic.AsyncMessageStream.__aiter__ = AsyncMessageStream__aiter__decorator(
        original_async_message_stream_aiter_method
    )

    async_message_stream_manager.opik_tracked_instance = True
    async_message_stream_manager.span_to_end = span_to_end
    async_message_stream_manager.trace_to_end = trace_to_end

    return async_message_stream_manager


def patch_sync_beta_message_stream_manager(
    message_stream_manager: BetaMessageStreamManager,
    span_to_end: span.SpanData,
    trace_to_end: Optional[trace.TraceData],
    finally_callback: generator_wrappers.FinishGeneratorCallback,
) -> BetaMessageStreamManager:
    """
    Patches `client.beta.messages.stream(...)` sync context manager.

    ```
    with client.beta.messages.stream(...) as stream:
        for event in stream:
            ...
    ```
    """

    def BetaMessageStream__iter__decorator(dunder_iter_func: Callable) -> Callable:
        @functools.wraps(dunder_iter_func)
        def wrapper(
            self: BetaMessageStream,
        ) -> Iterator[Any]:
            try:
                error_info: Optional[ErrorInfoDict] = None
                for item in dunder_iter_func(self):
                    yield item
            except Exception as exception:
                LOGGER.debug(
                    "Exception raised from anthropic.BetaMessageStream.",
                    exc_info=True,
                )
                error_info = error_info_collector.collect(exception)
                raise exception
            finally:
                if not hasattr(self, "opik_tracked_instance"):
                    return

                delattr(self, "opik_tracked_instance")

                accumulated_output = (
                    self.get_final_message() if error_info is None else None
                )

                finally_callback(
                    output=accumulated_output,
                    error_info=error_info,
                    capture_output=True,
                    generators_span_to_end=self.span_to_end,
                    generators_trace_to_end=self.trace_to_end,
                )

        return wrapper

    def BetaMessageStreamManager__enter__decorator(
        dunder_enter_func: Callable,
    ) -> Callable:
        @functools.wraps(dunder_enter_func)
        def wrapper(self: BetaMessageStreamManager) -> BetaMessageStream:
            result: BetaMessageStream = dunder_enter_func(self)

            if hasattr(self, "opik_tracked_instance"):
                result.opik_tracked_instance = True
                result.span_to_end = self.span_to_end
                result.trace_to_end = self.trace_to_end

            return result

        return wrapper

    BetaMessageStreamManager.__enter__ = BetaMessageStreamManager__enter__decorator(
        original_beta_message_stream_manager_enter_method
    )
    BetaMessageStream.__iter__ = BetaMessageStream__iter__decorator(
        original_beta_message_stream_iter_method
    )

    message_stream_manager.opik_tracked_instance = True
    message_stream_manager.span_to_end = span_to_end
    message_stream_manager.trace_to_end = trace_to_end

    return message_stream_manager


def patch_async_beta_message_stream_manager(
    async_message_stream_manager: BetaAsyncMessageStreamManager,
    span_to_end: span.SpanData,
    trace_to_end: Optional[trace.TraceData],
    finally_callback: generator_wrappers.FinishGeneratorCallback,
) -> BetaAsyncMessageStreamManager:
    """
    Patches `client.beta.messages.stream(...)` async context manager.

    ```
    async with client.beta.messages.stream(...) as stream:
        async for event in stream:
            ...
    ```
    """

    def BetaAsyncMessageStream__aiter__decorator(
        dunder_aiter_func: Callable,
    ) -> Callable:
        @functools.wraps(dunder_aiter_func)
        async def wrapper(
            self: BetaAsyncMessageStream,
        ) -> AsyncIterator[Any]:
            try:
                error_info: Optional[ErrorInfoDict] = None
                async for item in dunder_aiter_func(self):
                    yield item
            except Exception as exception:
                LOGGER.debug(
                    "Exception raised from anthropic.BetaAsyncMessageStream.",
                    exc_info=True,
                )
                error_info = error_info_collector.collect(exception)
                raise exception
            finally:
                if not hasattr(self, "opik_tracked_instance"):
                    return

                delattr(self, "opik_tracked_instance")

                accumulated_output = (
                    await self.get_final_message() if error_info is None else None
                )

                finally_callback(
                    output=accumulated_output,
                    error_info=error_info,
                    capture_output=True,
                    generators_span_to_end=self.span_to_end,
                    generators_trace_to_end=self.trace_to_end,
                )

        return wrapper

    def BetaAsyncMessageStreamManager__aenter__decorator(
        dunder_aenter_func: Callable,
    ) -> Callable:
        @functools.wraps(dunder_aenter_func)
        async def wrapper(
            self: BetaAsyncMessageStreamManager,
        ) -> BetaAsyncMessageStream:
            result: BetaAsyncMessageStream = await dunder_aenter_func(self)

            if hasattr(self, "opik_tracked_instance"):
                result.opik_tracked_instance = True
                result.span_to_end = self.span_to_end
                result.trace_to_end = self.trace_to_end

            return result

        return wrapper

    BetaAsyncMessageStreamManager.__aenter__ = (
        BetaAsyncMessageStreamManager__aenter__decorator(
            original_beta_async_message_stream_manager_aenter_method
        )
    )
    BetaAsyncMessageStream.__aiter__ = BetaAsyncMessageStream__aiter__decorator(
        original_beta_async_message_stream_aiter_method
    )

    async_message_stream_manager.opik_tracked_instance = True
    async_message_stream_manager.span_to_end = span_to_end
    async_message_stream_manager.trace_to_end = trace_to_end

    return async_message_stream_manager



import contextlib
from typing import Any, Dict, List, Optional, Iterator, Union

from opik import llm_usage
from opik.api_objects import span, trace, opik_client, prompt
from opik.api_objects.attachment import Attachment
from opik.types import (
    DistributedTraceHeadersDict,
    FeedbackScoreDict,
    LLMProvider,
    ErrorInfoDict,
)

from opik import tracing_runtime_config

from . import context_storage, exceptions
from .decorator import error_info_collector


def get_current_span_data() -> Optional[span.SpanData]:
    """
    Returns the current span created by track() decorator or None if no span was found.
    """
    span_data = context_storage.top_span_data()
    if span_data is None:
        return None

    return span.SpanData(**span_data.__dict__)


def get_current_trace_data() -> Optional[trace.TraceData]:
    """
    Returns the current trace created by track() decorator or None if no trace was found.
    """
    trace_data = context_storage.get_trace_data()
    if trace_data is None:
        return None

    return trace.TraceData(**trace_data.__dict__)


def get_distributed_trace_headers() -> DistributedTraceHeadersDict:
    """
    Returns headers' dictionary to be passed into tracked function on remote node.
    Requires an existing span in the context, otherwise raises an error.
    """
    current_span_data = context_storage.top_span_data()

    if current_span_data is None:
        raise exceptions.OpikException("There is no span in the context.")

    return DistributedTraceHeadersDict(
        opik_trace_id=current_span_data.trace_id,
        opik_parent_span_id=current_span_data.id,
    )


def update_current_span(
    name: Optional[str] = None,
    input: Optional[Dict[str, Any]] = None,
    output: Optional[Dict[str, Any]] = None,
    metadata: Optional[Dict[str, Any]] = None,
    tags: Optional[List[str]] = None,
    usage: Optional[Union[Dict[str, Any], llm_usage.OpikUsage]] = None,
    feedback_scores: Optional[List[FeedbackScoreDict]] = None,
    model: Optional[str] = None,
    provider: Optional[Union[str, LLMProvider]] = None,
    total_cost: Optional[float] = None,
    attachments: Optional[List[Attachment]] = None,
    error_info: Optional[ErrorInfoDict] = None,
    prompts: Optional[List[prompt.BasePrompt]] = None,
) -> None:
    """
    Update the current span with the provided parameters. This method is usually called within a tracked function.

    Args:
        name: The name of the span.
        input: The input data of the span.
        output: The output data of the span.
        metadata: The metadata of the span.
        tags: The tags of the span.
        usage: Usage data for the span. In order for input, output, and total tokens to be visible in the UI,
            the usage must contain OpenAI-formatted keys (they can be passed additionally to the original usage on the top level of the dict): prompt_tokens, completion_tokens, and total_tokens.
            If OpenAI-formatted keys were not found, Opik will try to calculate them automatically if the usage
            format is recognized (you can see which provider's formats are recognized in opik.LLMProvider enum), but it is not guaranteed.
        feedback_scores: The feedback scores of the span.
        model: The name of LLM (in this case type parameter should be == llm)
        provider: The provider of LLM. You can find providers officially supported by Opik for cost tracking
            in `opik.LLMProvider` enum. If your provider is not here, please open an issue in our GitHub - https://github.com/comet-ml/opik.
            If your provider is not in the list, you can still specify it, but the cost tracking will not be available
        total_cost: The cost of the span in USD. This value takes priority over the cost calculated by Opik from the usage.
        attachments: The list of attachments to be uploaded to the span.
        error_info: The error information of the span.
        prompts: The list of prompts used in the span.
    """
    if not tracing_runtime_config.is_tracing_active():
        return

    if prompts is not None:
        prompts = [p.__internal_api__to_info_dict__() for p in prompts]

    new_params = {
        "name": name,
        "input": input,
        "output": output,
        "metadata": metadata,
        "tags": tags,
        "usage": usage,
        "feedback_scores": feedback_scores,
        "model": model,
        "provider": provider,
        "total_cost": total_cost,
        "attachments": attachments,
        "error_info": error_info,
        "prompts": prompts,
    }
    current_span_data = context_storage.top_span_data()
    if current_span_data is None:
        raise exceptions.OpikException("There is no span in the context.")

    current_span_data.update(**new_params)


def update_current_trace(
    name: Optional[str] = None,
    input: Optional[Dict[str, Any]] = None,
    output: Optional[Dict[str, Any]] = None,
    metadata: Optional[Dict[str, Any]] = None,
    tags: Optional[List[str]] = None,
    feedback_scores: Optional[List[FeedbackScoreDict]] = None,
    thread_id: Optional[str] = None,
    attachments: Optional[List[Attachment]] = None,
    prompts: Optional[List[prompt.BasePrompt]] = None,
) -> None:
    """
    Update the current trace with the provided parameters. This method is usually called within a tracked function.

    Args:
        name: The name of the trace.
        input: The input data of the trace.
        output: The output data of the trace.
        metadata: The metadata of the trace.
        tags: The tags of the trace.
        feedback_scores: The feedback scores of the trace.
        thread_id: Used to group multiple traces into a thread.
            The identifier is user-defined and has to be unique per project.
        attachments: The list of attachments to be uploaded to the trace.
        prompts: The list of prompts used in the trace.
    """
    if not tracing_runtime_config.is_tracing_active():
        return

    if prompts is not None:
        prompts = [p.__internal_api__to_info_dict__() for p in prompts]

    new_params = {
        "name": name,
        "input": input,
        "output": output,
        "metadata": metadata,
        "tags": tags,
        "feedback_scores": feedback_scores,
        "thread_id": thread_id,
        "attachments": attachments,
        "prompts": prompts,
    }
    current_trace_data = context_storage.get_trace_data()
    if current_trace_data is None:
        raise exceptions.OpikException("There is no trace in the context.")

    current_trace_data.update(**new_params)


@contextlib.contextmanager
def trace_context(
    trace_data: trace.TraceData,
    client: opik_client.Opik,
) -> Iterator[None]:
    """
    Provides a context manager to handle trace data within an execution context.

    This function sets up trace data for the current context, ensuring it is
    properly cleaned up and processed during the lifecycle of the context. It also
    handles exceptions by collecting error information and associating it with
    the trace data before raising the exception further. At the end of the context,
    it finalizes the trace and logs it using the provided client.

    Args:
        trace_data: An instance of trace.TraceData containing information
            about the current trace context, such as start time, end time, and
            any relevant metadata for tracking execution.
        client: An object of type opik_client.Opik used to report the trace
            data, typically communicating with an external tracing or monitoring
            system.

    Yields:
        None: The context manager yields control back to the caller, allowing
            code execution within the defined trace context.

    Raises:
        Exception: The function raises any exceptions encountered within the
            context after collecting error information and associating it with the
            trace data.
    """
    if client.config.log_start_trace_span:
        client.__internal_api__trace__(**trace_data.as_start_parameters)

    error_info: Optional[ErrorInfoDict] = None
    try:
        context_storage.set_trace_data(trace_data)
        yield
    except Exception as exception:
        error_info = error_info_collector.collect(exception)
        raise
    finally:
        trace_data = context_storage.pop_trace_data()  # type: ignore

        assert trace_data is not None

        if error_info is not None:
            trace_data.error_info = error_info

        trace_data.init_end_time()

        client.__internal_api__trace__(**trace_data.as_parameters)


__all__ = [
    "get_current_span_data",
    "get_current_trace_data",
    "update_current_span",
    "update_current_trace",
    "get_distributed_trace_headers",
    "trace_context",
]



"""Usage report module for Opik CLI."""

from .cli import usage_report
from .extraction import extract_project_data
from .charts import create_charts, create_individual_chart
from .statistics import calculate_statistics
from .pdf import create_pdf_report

__all__ = [
    "usage_report",
    "extract_project_data",
    "create_charts",
    "create_individual_chart",
    "calculate_statistics",
    "create_pdf_report",
]



import random
from typing import List, Optional

from opik.api_objects.dataset import dataset_item

from . import base_dataset_sampler


class RandomDatasetSampler(base_dataset_sampler.BaseDatasetSampler):
    def __init__(
        self, max_samples: int, shuffle: bool = True, seed: Optional[int] = None
    ) -> None:
        """Samples a random subset of dataset items.

        This class is a dataset sampler that selects a random subset of items from a dataset.
        The number of items to sample can be specified, and shuffling can be enabled or disabled.
        An optional random seed can be provided for reproducibility.

        Args:
            max_samples: The maximum number of samples to generate.
            shuffle: Whether to shuffle the samples. Default is True, False provides a speedup
                for large datasets.
            seed: Seed for the random number generator. If None, then fresh, unpredictable
                entropy will be pulled from the OS.
        """
        self.max_samples = max_samples
        self.shuffle = shuffle
        self.seed = seed

    def sample(
        self, data_items: List[dataset_item.DatasetItem]
    ) -> List[dataset_item.DatasetItem]:
        if len(data_items) == 0:
            return []

        # Create a random number generator with the specified seed
        rng = random.Random(self.seed)

        # Determine how many samples to take
        sample_size = min(len(data_items), self.max_samples)

        # Do sample first to avoid shuffling the entire dataset
        items = rng.sample(data_items, sample_size)

        if self.shuffle:
            rng.shuffle(items)

        return items



import importlib.util
import logging
import sys
from typing import Optional, Any, Dict

from opik import config, _logging

from .litellm import litellm_chat_model
from . import base_model

LOGGER = logging.getLogger(__name__)

_MODEL_CACHE: Dict[Any, base_model.OpikBaseModel] = {}


def _freeze(value: Any) -> Any:
    if isinstance(value, dict):
        return frozenset((k, _freeze(v)) for k, v in value.items())
    if isinstance(value, (list, tuple)):
        return tuple(_freeze(v) for v in value)
    if isinstance(value, set):
        return frozenset(_freeze(v) for v in value)
    return value


def _make_cache_key(model_name: str, track: bool, model_kwargs: Dict[str, Any]) -> Any:
    frozen_kwargs = frozenset((k, _freeze(v)) for k, v in model_kwargs.items())
    return (model_name, track, frozen_kwargs)


def get_default_model_name() -> str:
    """Resolve the default model name from Opik config."""
    return config.OpikConfig().default_llm


def get(
    model_name: Optional[str], track: bool = True, **model_kwargs: Any
) -> base_model.OpikBaseModel:
    """
    Get or create a cached LiteLLM chat model instance.

    Args:
        model_name: The name of the model to use. Defaults to OpikConfig.default_llm if None.
        track: Whether to track the model calls. When False, disables tracing for this model instance.
            Defaults to True.
        **model_kwargs: Additional keyword arguments to pass to the model constructor.

    Returns:
        A cached or newly created OpikBaseModel instance.
    """
    if model_name is None:
        model_name = get_default_model_name()

    cache_key = _make_cache_key(model_name, track, model_kwargs)
    if cache_key not in _MODEL_CACHE:
        _MODEL_CACHE[cache_key] = _create_model(model_name, track, model_kwargs)
    return _MODEL_CACHE[cache_key]


def _is_anthropic_model(model_name: str) -> bool:
    return model_name.startswith("anthropic/") or model_name.startswith("claude")


def _should_use_anthropic_native(model_name: str) -> bool:
    if not _is_anthropic_model(model_name):
        return False
    if "anthropic" in sys.modules:
        return True
    if importlib.util.find_spec("anthropic") is not None:
        return True
    _logging.log_once_at_level(
        logging.WARNING,
        "Anthropic SDK is not installed. Falling back to LiteLLM for model '%s'. "
        "Install it with `pip install anthropic` for a more stable experience.",
        LOGGER,
        model_name,
    )
    return False


def _create_model(
    model_name: str, track: bool, model_kwargs: Dict[str, Any]
) -> base_model.OpikBaseModel:
    if _should_use_anthropic_native(model_name):
        LOGGER.debug(
            "Using native Anthropic SDK for model %s",
            model_name,
        )
        from .anthropic.anthropic_chat_model import AnthropicChatModel

        return AnthropicChatModel(model_name=model_name, track=track, **model_kwargs)

    return litellm_chat_model.LiteLLMChatModel(
        model_name=model_name, track=track, **model_kwargs
    )



import time
from enum import Enum
from typing import Optional

from opik.healthcheck import connection_probe


class ConnectionStatus(Enum):
    """
    Represents the connection status for a system.

    Provides an enumeration of various states a connection can be in, such as
    successful connection, connection failure, and restored connection.
    Used primarily to indicate the status of a network, server, or other connectivity-dependent services.

    Attributes:
        connection_ok: Indicates the connection is currently active and functioning properly.
        connection_failed: Indicates the connection has failed.
        connection_restored: Indicates the connection, which was previously failed, has been restored.
    """

    connection_ok = 1
    connection_failed = 2
    connection_restored = 3


class OpikConnectionMonitor:
    """
    Monitors the connection to the Opik server and tracks its status.

    This class is responsible for checking the server connectivity using a probe object
    at regular intervals. It maintains the status of the server connection, including
    disconnection details and reasons.

    Attributes:
        ping_interval: Interval in seconds between server connection pings.
        disconnect_time: Time in seconds when the server connection was lost.
            Defaults to 0 if the connection has not been lost yet.
        disconnect_reason: Reason for the disconnection, if applicable.
            Defaults to None if there is no disconnection.
    """

    def __init__(
        self,
        ping_interval: float,
        check_timeout: float,
        probe: connection_probe.ConnectionProbe,
    ):
        """
        Initializes an instance of the class to handle connection probing and monitoring.

        Args:
            ping_interval: The interval, in seconds, at which pings are sent to probe
                the connection.
            check_timeout: The threshold duration, in seconds, before a connection is
                considered lost if no response is received.
            probe: An instance of the ``connection_probe.ConnectionProbe`` class used for
                sending probe requests and handling connectivity checks.
        """
        self._check_timeout = check_timeout
        self._has_server_connection = True
        self._probe = probe

        self.last_beat = float("-inf")
        self.ping_interval = ping_interval
        self.disconnect_time = 0.0
        self.disconnect_reason: Optional[str] = None

    @property
    def has_server_connection(self) -> bool:
        return self._has_server_connection

    def reset(self) -> None:
        self.disconnect_time = 0.0
        self.disconnect_reason = None
        self._has_server_connection = True

    def connection_failed(self, failure_reason: Optional[str]) -> None:
        """
        Handles the procedure when the connection to the server is determined to have failed.

        This method processes the failure by capturing the disconnection time and the reason for the failure
        (if provided). It also updates the state to reflect that the connection to the server is no longer active.

        Args:
            failure_reason: Optional reason for the connection failure. Defaults to None if no reason
                is provided.
        """
        if self._has_server_connection:
            # save the first disconnection time and reason
            self.disconnect_time = time.time()
            self.disconnect_reason = failure_reason

        self._has_server_connection = False

    def tick(self) -> ConnectionStatus:
        """
        Checks and updates the connection status during a timed heartbeat event.

        This method ensures that the connection is actively monitored, sends probe
        requests to check the connection status if necessary, and updates the last
        heartbeat timestamp.

        Returns:
            ConnectionStatus: Indicates the current status of the connection. It may
            return one of the following statuses:
            - `ConnectionStatus.connection_ok`: The connection is active.
            - `ConnectionStatus.connection_failed`: The connection is not active.
            - Result based on the probe check indicating whether the connection is
              healthy or not.
        """
        next_beat = self.last_beat + self.ping_interval
        now = time.time()
        if next_beat <= now:
            result = self._probe.check_connection(timeout=self._check_timeout)
            self.last_beat = time.time()
            return self._on_ping_result(
                result.is_healthy, failure_reason=result.error_message
            )
        elif self.has_server_connection:
            return ConnectionStatus.connection_ok
        else:
            return ConnectionStatus.connection_failed

    def _on_ping_result(
        self, success: bool, failure_reason: Optional[str]
    ) -> ConnectionStatus:
        """Invoked to process a received ping result"""
        if success:
            if not self.has_server_connection:
                status = ConnectionStatus.connection_restored
            else:
                status = ConnectionStatus.connection_ok

            self._has_server_connection = True
        else:
            status = ConnectionStatus.connection_failed
            self.connection_failed(failure_reason)

        return status



import logging
import threading
import time
from collections import defaultdict
from concurrent import futures
from typing import Any, Dict, List, Optional, TypeVar, Generic

from ..metrics.score_result import ScoreResult
from .types import EvaluationTask

LOGGER = logging.getLogger(__name__)

T = TypeVar("T")


class StreamingExecutor(Generic[T]):
    """
    Executor that accepts and processes evaluation tasks incrementally using a thread pool.

    Tasks can be submitted one at a time and will begin executing immediately, allowing
    for streaming behavior regardless of the number of workers configured.

    Progress bar updates happen immediately as tasks complete via future callbacks,
    even while the main thread is still submitting new tasks.
    """

    def __init__(
        self,
        workers: int,
        verbose: int,
        desc: str = "Evaluation",
        total: Optional[int] = None,
        show_score_postfix: bool = True,
    ):
        self._workers = workers
        self._verbose = verbose
        self._desc = desc
        self._total = total
        self._show_score_postfix = show_score_postfix
        self._task_count = 0
        self._pool: futures.ThreadPoolExecutor
        self._submitted_futures: List[futures.Future[T]] = []
        self._progress_bar: Optional[Any] = None
        self._future_to_group: Dict[futures.Future[T], str] = {}
        self._group_sizes: Dict[str, int] = {}

        # Callback state — accessed from worker threads under lock
        self._lock = threading.Lock()
        self._results: List[T] = []
        self._score_totals: Dict[str, float] = defaultdict(float)
        self._score_counts: Dict[str, int] = defaultdict(int)
        self._group_completed: Dict[str, int] = defaultdict(int)
        self._first_update_done = False

    def __enter__(self) -> "StreamingExecutor[T]":
        self._pool = futures.ThreadPoolExecutor(max_workers=self._workers)
        self._pool.__enter__()
        # Initialize progress bar on enter (lazy import for mockability)
        from opik.environment import get_tqdm_for_current_environment

        _tqdm = get_tqdm_for_current_environment()
        self._progress_bar = _tqdm(
            disable=(self._verbose < 1),
            desc=f"{self._desc} (there might be a delay before the first items are processed)",
            total=self._total,
        )
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        # Close progress bar if it exists
        if self._progress_bar is not None:
            self._progress_bar.close()
        self._pool.__exit__(exc_type, exc_val, exc_tb)

    def set_group_size(self, group_id: str, size: int) -> None:
        """Declare the expected number of tasks for a group.

        Must be called before submitting tasks for that group to avoid
        race conditions between task completion callbacks and submission.
        """
        self._group_sizes[group_id] = size

    def submit(self, task: EvaluationTask[T], group_id: Optional[str] = None) -> None:
        """Submit a task to the thread pool for execution.

        Args:
            task: The evaluation task to execute.
            group_id: Optional group identifier. When provided, progress bar
                updates once per group (when all tasks in the group complete)
                instead of once per task. The group size must be declared
                via ``set_group_size()`` before submitting grouped tasks.
        """
        self._task_count += 1
        future = self._pool.submit(task)
        self._submitted_futures.append(future)

        if group_id is not None:
            self._future_to_group[future] = group_id

        future.add_done_callback(self._on_future_done)

    def _on_future_done(self, future: futures.Future[T]) -> None:
        """Callback fired by worker thread when a task completes."""
        exc = future.exception()
        if exc is not None:
            group_id = self._future_to_group.get(future)
            LOGGER.warning(
                "Evaluation task failed (group=%s): %s: %s",
                group_id,
                type(exc).__name__,
                exc,
            )
            return

        result = future.result()

        with self._lock:
            self._results.append(result)

            # Update running scores if result has score_results attribute
            if hasattr(result, "score_results") and isinstance(
                result.score_results, list
            ):
                for score in result.score_results:
                    if isinstance(score, ScoreResult) and not score.scoring_failed:
                        self._score_totals[score.name] += score.value
                        self._score_counts[score.name] += 1

                # Update progress bar with running averages
                if (
                    self._progress_bar is not None
                    and self._score_counts
                    and self._show_score_postfix
                ):
                    postfix_dict = {
                        name: f"{self._score_totals[name] / self._score_counts[name]:.4f}"
                        for name in self._score_counts
                    }
                    self._progress_bar.set_postfix(postfix_dict)

            # Update progress bar
            if self._progress_bar is not None:
                should_update = False
                if future in self._future_to_group:
                    gid = self._future_to_group[future]
                    self._group_completed[gid] += 1
                    if self._group_completed[gid] == self._group_sizes[gid]:
                        should_update = True
                else:
                    should_update = True

                if should_update:
                    if not self._first_update_done:
                        self._first_update_done = True
                        self._progress_bar.set_description(self._desc)
                    self._progress_bar.update(1)

    def get_results(self) -> List[T]:
        """Wait for all submitted tasks and return their results.

        Progress tracking happens via callbacks during both the submit and
        wait phases. This method sets the total if it wasn't known at
        construction, waits for completion, and re-raises any exceptions.
        """
        # Update total if it wasn't known initially
        if self._progress_bar is not None and self._total is None:
            use_groups = bool(self._future_to_group)
            if use_groups:
                self._progress_bar.total = len(self._group_sizes)
            else:
                self._progress_bar.total = self._task_count

        # Wait for all futures to complete
        LOGGER.debug(
            "[executor] Waiting for %d futures to complete...",
            len(self._submitted_futures),
        )
        wait_start = time.monotonic()
        futures.wait(self._submitted_futures)
        elapsed = time.monotonic() - wait_start
        LOGGER.debug(
            "[executor] All %d futures completed in %.1fs",
            len(self._submitted_futures),
            elapsed,
        )

        # Re-raise first exception if any task failed
        for future in self._submitted_futures:
            future.result()

        with self._lock:
            return list(self._results)


def execute(
    evaluation_tasks: List[EvaluationTask[T]],
    workers: int,
    verbose: int,
    desc: str = "Evaluation",
) -> List[T]:
    if workers == 1:
        from opik.environment import get_tqdm_for_current_environment

        _tqdm = get_tqdm_for_current_environment()
        test_results = [
            evaluation_task()
            for evaluation_task in _tqdm(
                evaluation_tasks,
                disable=(verbose < 1),
                desc=desc,
                total=len(evaluation_tasks),
            )
        ]

        return test_results

    with StreamingExecutor[T](
        workers=workers, verbose=verbose, desc=desc, total=len(evaluation_tasks)
    ) as executor:
        for evaluation_task in evaluation_tasks:
            executor.submit(evaluation_task)
        return executor.get_results()



import importlib.util
import logging
from typing import Any, Dict, List, Optional, Type

import pydantic
import tenacity

import opik.config as opik_config
from .. import base_model
from opik import exceptions
from . import message_adapter, response_parser

LOGGER = logging.getLogger(__name__)

DEFAULT_MAX_TOKENS = 4096


def is_available() -> bool:
    return importlib.util.find_spec("anthropic") is not None


class AnthropicChatModel(base_model.OpikBaseModel):
    def __init__(
        self,
        model_name: str = "anthropic/claude-sonnet-4-20250514",
        track: bool = True,
        **completion_kwargs: Any,
    ) -> None:
        """
        Initializes the Anthropic model using the native anthropic Python SDK.

        Args:
            model_name: The name of the model. Use "anthropic/" prefix or bare model name.
            track: Whether to track the model calls via Opik tracing.
            completion_kwargs: Additional arguments always passed to messages.create().
        """
        import anthropic

        super().__init__(model_name=model_name)

        self._api_model_name = message_adapter.strip_anthropic_prefix(model_name)
        self._unsupported_warned: set[str] = set()
        self._completion_kwargs = message_adapter.filter_unsupported_params(
            completion_kwargs, self._unsupported_warned
        )

        config = opik_config.OpikConfig()
        enable_tracking = track and config.enable_litellm_models_monitoring

        client = anthropic.Anthropic(timeout=60.0)
        async_client = anthropic.AsyncAnthropic(timeout=60.0)

        if enable_tracking:
            from opik.integrations.anthropic import track_anthropic

            client = track_anthropic(client)
            async_client = track_anthropic(async_client)

        self._client = client
        self._async_client = async_client

    def generate_string(
        self,
        input: str,
        response_format: Optional[Type[pydantic.BaseModel]] = None,
        **kwargs: Any,
    ) -> str:
        request = [{"content": input, "role": "user"}]

        if response_format is not None:
            kwargs["response_format"] = response_format

        with base_model.get_provider_response(
            model_provider=self,
            messages=request,
            **kwargs,
        ) as response:
            content = response_parser.extract_text_content(response)
            return base_model.check_model_output_string(content)

    def generate_provider_response(
        self,
        messages: List[Dict[str, Any]],
        **kwargs: Any,
    ) -> Any:
        retries = kwargs.pop("__opik_retries", 3)
        try:
            max_attempts = max(1, int(retries))
        except (TypeError, ValueError):
            max_attempts = 1

        response_format = kwargs.pop("response_format", None)
        call_kwargs = self._build_call_kwargs(messages, kwargs)

        retrying = tenacity.Retrying(
            reraise=True,
            stop=tenacity.stop_after_attempt(max_attempts),
            wait=tenacity.wait_exponential(multiplier=0.5, min=0.5, max=8.0),
        )

        if response_format is not None:
            call_kwargs["output_format"] = response_format
            return retrying(self._client.messages.parse, **call_kwargs)

        return retrying(self._client.messages.create, **call_kwargs)

    async def agenerate_string(
        self,
        input: str,
        response_format: Optional[Type[pydantic.BaseModel]] = None,
        **kwargs: Any,
    ) -> str:
        request = [{"content": input, "role": "user"}]

        if response_format is not None:
            kwargs["response_format"] = response_format

        async with base_model.aget_provider_response(
            model_provider=self, messages=request, **kwargs
        ) as response:
            content = response_parser.extract_text_content(response)
            return base_model.check_model_output_string(content)

    async def agenerate_provider_response(
        self,
        messages: List[Dict[str, Any]],
        **kwargs: Any,
    ) -> Any:
        retries = kwargs.pop("__opik_retries", 3)
        try:
            max_attempts = max(1, int(retries))
        except (TypeError, ValueError):
            max_attempts = 1

        response_format = kwargs.pop("response_format", None)
        call_kwargs = self._build_call_kwargs(messages, kwargs)

        retrying = tenacity.AsyncRetrying(
            reraise=True,
            stop=tenacity.stop_after_attempt(max_attempts),
            wait=tenacity.wait_exponential(multiplier=0.5, min=0.5, max=8.0),
        )

        if response_format is not None:
            call_kwargs["output_format"] = response_format
            async for attempt in retrying:
                with attempt:
                    return await self._async_client.messages.parse(**call_kwargs)
        else:
            async for attempt in retrying:
                with attempt:
                    return await self._async_client.messages.create(**call_kwargs)

        raise exceptions.BaseLLMError(
            "Async Anthropic completion failed without raising an exception"
        )

    def _build_call_kwargs(
        self,
        messages: List[Dict[str, Any]],
        extra_kwargs: Dict[str, Any],
    ) -> Dict[str, Any]:
        system_text, non_system_messages = message_adapter.extract_system_messages(
            messages
        )

        filtered_extra = message_adapter.filter_unsupported_params(
            extra_kwargs, self._unsupported_warned
        )

        call_kwargs: Dict[str, Any] = {
            **self._completion_kwargs,
            **filtered_extra,
        }
        call_kwargs["model"] = self._api_model_name
        call_kwargs["messages"] = non_system_messages
        call_kwargs.setdefault("max_tokens", DEFAULT_MAX_TOKENS)

        if system_text:
            call_kwargs["system"] = system_text

        return call_kwargs



from .text.prompt import Prompt
from .text.prompt_template import PromptTemplate
from .chat.chat_prompt import ChatPrompt
from .chat.chat_prompt_template import ChatPromptTemplate
from .types import PromptType
from .base_prompt import BasePrompt
from .base_prompt_template import BasePromptTemplate

__all__ = [
    "PromptType",
    "Prompt",
    "ChatPrompt",
    "PromptTemplate",
    "ChatPromptTemplate",
    "BasePrompt",
    "BasePromptTemplate",
]



from typing import List, Dict

from . import nodes, subgraph_edges_builders
import google.adk.agents
import functools

CLASS_STYLES: Dict[nodes.GraphNodeType, str] = {
    nodes.GraphNodeType.SEQUENTIAL_AGENT: "fill:#d9f2d9,stroke:#339933",
    nodes.GraphNodeType.LOOP_AGENT: "fill:#e6ccff,stroke:#6600cc",
    nodes.GraphNodeType.PARALLEL_AGENT: "fill:#ffcccc,stroke:#cc0000",
    nodes.GraphNodeType.LLM_AGENT: "fill:#b3e0ff,stroke:#0077b3",
    nodes.GraphNodeType.TOOL: "fill:#ffcc99,stroke:#ff8000",
}

NODE_TYPES_TO_BUILD_SUBGRAPHS = [
    nodes.GraphNodeType.SEQUENTIAL_AGENT,
    nodes.GraphNodeType.LOOP_AGENT,
    nodes.GraphNodeType.PARALLEL_AGENT,
]


class MermaidGraphBuilder:
    def __init__(self, root_name: str):
        self.subgraphs_definitions: List[str] = []
        self.edges_definitions: List[str] = []
        self.style_definitions: List[str] = []
        self.root_name = root_name

    def build_node_graph(self, node: nodes.AgentNode) -> None:
        # 1. Create subgraph for composite nodes or process
        # connections for non-composite nodes with subagents
        if node.type in NODE_TYPES_TO_BUILD_SUBGRAPHS:
            self._build_subgraph_for_composite_node(node)
        elif len(node.subagent_nodes) > 0 or len(node.tools) > 0:
            self._build_edges_for_non_composite_llm_node(node)

        self._add_style(node.name, node.type)

        # 2. Recursively process all subagents
        for child in node.subagent_nodes:
            self.build_node_graph(child)

        # 3. Recursively process all tools
        for tool in node.tools:
            self._add_style(tool.name, nodes.GraphNodeType.TOOL)
            if tool.agent is not None:
                self.edges_definitions.append(f"{tool.name} --> {tool.agent.name}")
                self.build_node_graph(tool.agent)

    def _build_edges_for_non_composite_llm_node(self, node: nodes.AgentNode) -> None:
        for child in node.subagent_nodes:
            self.edges_definitions.append(f"{node.name} --> {child.name}")

        for tool in node.tools:
            self.edges_definitions.append(f"{node.name} --> {tool.name}")

    def _build_subgraph_for_composite_node(
        self, composite_node: nodes.AgentNode
    ) -> None:
        block = [f'subgraph {composite_node.name}["{composite_node.name}"]']

        if composite_node.type == nodes.GraphNodeType.SEQUENTIAL_AGENT:
            edge_definitions = (
                subgraph_edges_builders.build_edge_definitions_for_sequential_subagents(
                    composite_node.subagent_nodes
                )
            )
            block.extend([f"  {edge}" for edge in edge_definitions])

        elif composite_node.type == nodes.GraphNodeType.LOOP_AGENT:
            edge_definitions = (
                subgraph_edges_builders.build_edge_definitions_for_loop_subagents(
                    composite_node.subagent_nodes
                )
            )
            block.extend([f"  {edge}" for edge in edge_definitions])

        elif composite_node.type == nodes.GraphNodeType.PARALLEL_AGENT:
            edge_definitions = (
                subgraph_edges_builders.build_edge_definitions_for_parallel_subagents(
                    composite_node.subagent_nodes
                )
            )
            block.extend([f"  {edge}" for edge in edge_definitions])

        block.append("end")
        self.subgraphs_definitions.append("\n".join(block))

    @functools.lru_cache
    def _add_style(self, node_name: str, node_type: nodes.GraphNodeType) -> None:
        if node_type in CLASS_STYLES:
            self.style_definitions.append(
                f"style {node_name} {CLASS_STYLES[node_type]}"
            )

    def get_mermaid_source(self) -> str:
        return "\n".join(
            ["flowchart LR", f'{self.root_name}["{self.root_name}"]']
            + self.subgraphs_definitions
            + self.edges_definitions
            + self.style_definitions
        )


def build_mermaid_graph_definition(root_agent: google.adk.agents.BaseAgent) -> str:
    """
    Generates a Mermaid 'flowchart LR' diagram for a google-adk
    agent tree and returns the Mermaid source code.

    Args:
        root_agent (Agent):
            The root agent node of the google-adk agent tree.
            This should be an instance
            of SequentialAgent, LoopAgent, ParallelAgent or LlmAgent.

    Returns:
        str: The Mermaid source code as a string.

    Example:
        >>> mermaid_src = build_mermaid_graph_definition(my_agent_tree)
        >>> print(mermaid_src)
    """
    parsed_agent_tree = nodes.build_nodes_tree(root_agent)

    graph_builder = MermaidGraphBuilder(root_name=root_agent.name)
    graph_builder.build_node_graph(parsed_agent_tree)

    return graph_builder.get_mermaid_source()



from opik.types import ErrorInfoDict
import traceback


def collect(exception: Exception) -> ErrorInfoDict:
    result: ErrorInfoDict = {
        "exception_type": type(exception).__name__,
        "traceback": "".join(
            traceback.TracebackException.from_exception(exception).format()
        ),
    }

    message = str(exception)
    if message != "":
        result["message"] = message

    return result



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .project_metric_response_public_interval import ProjectMetricResponsePublicInterval
from .project_metric_response_public_metric_type import ProjectMetricResponsePublicMetricType
from .results_number_public import ResultsNumberPublic


class ProjectMetricResponsePublic(UniversalBaseModel):
    project_id: typing.Optional[str] = None
    metric_type: typing.Optional[ProjectMetricResponsePublicMetricType] = None
    interval: typing.Optional[ProjectMetricResponsePublicInterval] = None
    results: typing.Optional[typing.List[ResultsNumberPublic]] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.http_response import AsyncHttpResponse, HttpResponse
from ..core.jsonable_encoder import jsonable_encoder
from ..core.pydantic_utilities import parse_obj_as
from ..core.request_options import RequestOptions
from ..core.serialization import convert_and_respect_annotation_metadata
from ..errors.bad_request_error import BadRequestError
from ..errors.conflict_error import ConflictError
from ..errors.not_found_error import NotFoundError
from ..errors.unprocessable_entity_error import UnprocessableEntityError
from ..types.alert_page_public import AlertPagePublic
from ..types.alert_public import AlertPublic
from ..types.alert_trigger_write import AlertTriggerWrite
from ..types.alert_write_alert_type import AlertWriteAlertType
from ..types.webhook_examples import WebhookExamples
from ..types.webhook_test_result import WebhookTestResult
from ..types.webhook_write import WebhookWrite
from .types.get_webhook_examples_request_alert_type import GetWebhookExamplesRequestAlertType

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawAlertsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def find_alerts(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        sorting: typing.Optional[str] = None,
        filters: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[AlertPagePublic]:
        """
        Find alerts

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        sorting : typing.Optional[str]

        filters : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[AlertPagePublic]
            OK
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/alerts",
            method="GET",
            params={
                "page": page,
                "size": size,
                "sorting": sorting,
                "filters": filters,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AlertPagePublic,
                    parse_obj_as(
                        type_=AlertPagePublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def create_alert(
        self,
        *,
        webhook: WebhookWrite,
        id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        enabled: typing.Optional[bool] = OMIT,
        alert_type: typing.Optional[AlertWriteAlertType] = OMIT,
        metadata: typing.Optional[typing.Dict[str, str]] = OMIT,
        triggers: typing.Optional[typing.Sequence[AlertTriggerWrite]] = OMIT,
        project_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Create alert

        Parameters
        ----------
        webhook : WebhookWrite

        id : typing.Optional[str]

        name : typing.Optional[str]

        enabled : typing.Optional[bool]

        alert_type : typing.Optional[AlertWriteAlertType]

        metadata : typing.Optional[typing.Dict[str, str]]

        triggers : typing.Optional[typing.Sequence[AlertTriggerWrite]]

        project_id : typing.Optional[str]
            Optional project scope for this alert. When set, the alert is scoped to the specified project. Do NOT also provide a 'scope:project' trigger config — the system will create it automatically from this field. Sending both project_id and a scope:project trigger config will result in an error.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/alerts",
            method="POST",
            json={
                "id": id,
                "name": name,
                "enabled": enabled,
                "alert_type": alert_type,
                "metadata": metadata,
                "webhook": convert_and_respect_annotation_metadata(
                    object_=webhook, annotation=WebhookWrite, direction="write"
                ),
                "triggers": convert_and_respect_annotation_metadata(
                    object_=triggers, annotation=typing.Sequence[AlertTriggerWrite], direction="write"
                ),
                "project_id": project_id,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 409:
                raise ConflictError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def delete_alert_batch(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[None]:
        """
        Delete multiple alerts by their IDs

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/alerts/delete",
            method="POST",
            json={
                "ids": ids,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_alert_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[AlertPublic]:
        """
        Get Alert by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[AlertPublic]
            Alert resource
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/alerts/{jsonable_encoder(id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AlertPublic,
                    parse_obj_as(
                        type_=AlertPublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def update_alert(
        self,
        id_: str,
        *,
        webhook: WebhookWrite,
        id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        enabled: typing.Optional[bool] = OMIT,
        alert_type: typing.Optional[AlertWriteAlertType] = OMIT,
        metadata: typing.Optional[typing.Dict[str, str]] = OMIT,
        triggers: typing.Optional[typing.Sequence[AlertTriggerWrite]] = OMIT,
        project_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Update alert

        Parameters
        ----------
        id_ : str

        webhook : WebhookWrite

        id : typing.Optional[str]

        name : typing.Optional[str]

        enabled : typing.Optional[bool]

        alert_type : typing.Optional[AlertWriteAlertType]

        metadata : typing.Optional[typing.Dict[str, str]]

        triggers : typing.Optional[typing.Sequence[AlertTriggerWrite]]

        project_id : typing.Optional[str]
            Optional project scope for this alert. When set, the alert is scoped to the specified project. Do NOT also provide a 'scope:project' trigger config — the system will create it automatically from this field. Sending both project_id and a scope:project trigger config will result in an error.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/alerts/{jsonable_encoder(id_)}",
            method="PUT",
            json={
                "id": id,
                "name": name,
                "enabled": enabled,
                "alert_type": alert_type,
                "metadata": metadata,
                "webhook": convert_and_respect_annotation_metadata(
                    object_=webhook, annotation=WebhookWrite, direction="write"
                ),
                "triggers": convert_and_respect_annotation_metadata(
                    object_=triggers, annotation=typing.Sequence[AlertTriggerWrite], direction="write"
                ),
                "project_id": project_id,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 409:
                raise ConflictError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_webhook_examples(
        self,
        *,
        alert_type: typing.Optional[GetWebhookExamplesRequestAlertType] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[WebhookExamples]:
        """
        Get webhook payload examples for all alert event types, optionally filtered by alert type

        Parameters
        ----------
        alert_type : typing.Optional[GetWebhookExamplesRequestAlertType]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[WebhookExamples]
            Webhook examples
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/alerts/webhooks/examples",
            method="GET",
            params={
                "alert_type": alert_type,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    WebhookExamples,
                    parse_obj_as(
                        type_=WebhookExamples,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def test_webhook(
        self,
        *,
        webhook: WebhookWrite,
        id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        enabled: typing.Optional[bool] = OMIT,
        alert_type: typing.Optional[AlertWriteAlertType] = OMIT,
        metadata: typing.Optional[typing.Dict[str, str]] = OMIT,
        triggers: typing.Optional[typing.Sequence[AlertTriggerWrite]] = OMIT,
        project_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[WebhookTestResult]:
        """
        Test alert webhook

        Parameters
        ----------
        webhook : WebhookWrite

        id : typing.Optional[str]

        name : typing.Optional[str]

        enabled : typing.Optional[bool]

        alert_type : typing.Optional[AlertWriteAlertType]

        metadata : typing.Optional[typing.Dict[str, str]]

        triggers : typing.Optional[typing.Sequence[AlertTriggerWrite]]

        project_id : typing.Optional[str]
            Optional project scope for this alert. When set, the alert is scoped to the specified project. Do NOT also provide a 'scope:project' trigger config — the system will create it automatically from this field. Sending both project_id and a scope:project trigger config will result in an error.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[WebhookTestResult]
            Webhook test
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/alerts/webhooks/tests",
            method="POST",
            json={
                "id": id,
                "name": name,
                "enabled": enabled,
                "alert_type": alert_type,
                "metadata": metadata,
                "webhook": convert_and_respect_annotation_metadata(
                    object_=webhook, annotation=WebhookWrite, direction="write"
                ),
                "triggers": convert_and_respect_annotation_metadata(
                    object_=triggers, annotation=typing.Sequence[AlertTriggerWrite], direction="write"
                ),
                "project_id": project_id,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    WebhookTestResult,
                    parse_obj_as(
                        type_=WebhookTestResult,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)


class AsyncRawAlertsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def find_alerts(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        sorting: typing.Optional[str] = None,
        filters: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[AlertPagePublic]:
        """
        Find alerts

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        sorting : typing.Optional[str]

        filters : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[AlertPagePublic]
            OK
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/alerts",
            method="GET",
            params={
                "page": page,
                "size": size,
                "sorting": sorting,
                "filters": filters,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AlertPagePublic,
                    parse_obj_as(
                        type_=AlertPagePublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def create_alert(
        self,
        *,
        webhook: WebhookWrite,
        id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        enabled: typing.Optional[bool] = OMIT,
        alert_type: typing.Optional[AlertWriteAlertType] = OMIT,
        metadata: typing.Optional[typing.Dict[str, str]] = OMIT,
        triggers: typing.Optional[typing.Sequence[AlertTriggerWrite]] = OMIT,
        project_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Create alert

        Parameters
        ----------
        webhook : WebhookWrite

        id : typing.Optional[str]

        name : typing.Optional[str]

        enabled : typing.Optional[bool]

        alert_type : typing.Optional[AlertWriteAlertType]

        metadata : typing.Optional[typing.Dict[str, str]]

        triggers : typing.Optional[typing.Sequence[AlertTriggerWrite]]

        project_id : typing.Optional[str]
            Optional project scope for this alert. When set, the alert is scoped to the specified project. Do NOT also provide a 'scope:project' trigger config — the system will create it automatically from this field. Sending both project_id and a scope:project trigger config will result in an error.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/alerts",
            method="POST",
            json={
                "id": id,
                "name": name,
                "enabled": enabled,
                "alert_type": alert_type,
                "metadata": metadata,
                "webhook": convert_and_respect_annotation_metadata(
                    object_=webhook, annotation=WebhookWrite, direction="write"
                ),
                "triggers": convert_and_respect_annotation_metadata(
                    object_=triggers, annotation=typing.Sequence[AlertTriggerWrite], direction="write"
                ),
                "project_id": project_id,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 409:
                raise ConflictError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def delete_alert_batch(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[None]:
        """
        Delete multiple alerts by their IDs

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/alerts/delete",
            method="POST",
            json={
                "ids": ids,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_alert_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[AlertPublic]:
        """
        Get Alert by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[AlertPublic]
            Alert resource
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/alerts/{jsonable_encoder(id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    AlertPublic,
                    parse_obj_as(
                        type_=AlertPublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def update_alert(
        self,
        id_: str,
        *,
        webhook: WebhookWrite,
        id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        enabled: typing.Optional[bool] = OMIT,
        alert_type: typing.Optional[AlertWriteAlertType] = OMIT,
        metadata: typing.Optional[typing.Dict[str, str]] = OMIT,
        triggers: typing.Optional[typing.Sequence[AlertTriggerWrite]] = OMIT,
        project_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Update alert

        Parameters
        ----------
        id_ : str

        webhook : WebhookWrite

        id : typing.Optional[str]

        name : typing.Optional[str]

        enabled : typing.Optional[bool]

        alert_type : typing.Optional[AlertWriteAlertType]

        metadata : typing.Optional[typing.Dict[str, str]]

        triggers : typing.Optional[typing.Sequence[AlertTriggerWrite]]

        project_id : typing.Optional[str]
            Optional project scope for this alert. When set, the alert is scoped to the specified project. Do NOT also provide a 'scope:project' trigger config — the system will create it automatically from this field. Sending both project_id and a scope:project trigger config will result in an error.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/alerts/{jsonable_encoder(id_)}",
            method="PUT",
            json={
                "id": id,
                "name": name,
                "enabled": enabled,
                "alert_type": alert_type,
                "metadata": metadata,
                "webhook": convert_and_respect_annotation_metadata(
                    object_=webhook, annotation=WebhookWrite, direction="write"
                ),
                "triggers": convert_and_respect_annotation_metadata(
                    object_=triggers, annotation=typing.Sequence[AlertTriggerWrite], direction="write"
                ),
                "project_id": project_id,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 409:
                raise ConflictError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_webhook_examples(
        self,
        *,
        alert_type: typing.Optional[GetWebhookExamplesRequestAlertType] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[WebhookExamples]:
        """
        Get webhook payload examples for all alert event types, optionally filtered by alert type

        Parameters
        ----------
        alert_type : typing.Optional[GetWebhookExamplesRequestAlertType]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[WebhookExamples]
            Webhook examples
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/alerts/webhooks/examples",
            method="GET",
            params={
                "alert_type": alert_type,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    WebhookExamples,
                    parse_obj_as(
                        type_=WebhookExamples,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def test_webhook(
        self,
        *,
        webhook: WebhookWrite,
        id: typing.Optional[str] = OMIT,
        name: typing.Optional[str] = OMIT,
        enabled: typing.Optional[bool] = OMIT,
        alert_type: typing.Optional[AlertWriteAlertType] = OMIT,
        metadata: typing.Optional[typing.Dict[str, str]] = OMIT,
        triggers: typing.Optional[typing.Sequence[AlertTriggerWrite]] = OMIT,
        project_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[WebhookTestResult]:
        """
        Test alert webhook

        Parameters
        ----------
        webhook : WebhookWrite

        id : typing.Optional[str]

        name : typing.Optional[str]

        enabled : typing.Optional[bool]

        alert_type : typing.Optional[AlertWriteAlertType]

        metadata : typing.Optional[typing.Dict[str, str]]

        triggers : typing.Optional[typing.Sequence[AlertTriggerWrite]]

        project_id : typing.Optional[str]
            Optional project scope for this alert. When set, the alert is scoped to the specified project. Do NOT also provide a 'scope:project' trigger config — the system will create it automatically from this field. Sending both project_id and a scope:project trigger config will result in an error.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[WebhookTestResult]
            Webhook test
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/alerts/webhooks/tests",
            method="POST",
            json={
                "id": id,
                "name": name,
                "enabled": enabled,
                "alert_type": alert_type,
                "metadata": metadata,
                "webhook": convert_and_respect_annotation_metadata(
                    object_=webhook, annotation=WebhookWrite, direction="write"
                ),
                "triggers": convert_and_respect_annotation_metadata(
                    object_=triggers, annotation=typing.Sequence[AlertTriggerWrite], direction="write"
                ),
                "project_id": project_id,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    WebhookTestResult,
                    parse_obj_as(
                        type_=WebhookTestResult,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)



import abc
import logging
from typing import List, Optional, TypeVar, Type

from .. import messages

import opik.exceptions


LOGGER = logging.getLogger(__name__)

T = TypeVar("T")


class BaseMessageProcessor(abc.ABC):
    @abc.abstractmethod
    def process(self, message: messages.BaseMessage) -> None:
        pass

    @abc.abstractmethod
    def is_active(self) -> bool:
        return False


class ChainedMessageProcessor(BaseMessageProcessor):
    """
    Processes messages through a chain of message processors.

    This class allows for the sequential processing of a message by a list of
    `BaseMessageProcessor` instances. Each processor in the chain is invoked in the
    order provided. If an exception occurs during the processing by a specific
    processor, it is logged, and the process continues with the next processor in
    the chain.
    """

    def __init__(self, processors: List[BaseMessageProcessor]) -> None:
        self._processors = processors

    def is_active(self) -> bool:
        return True

    def process(self, message: messages.BaseMessage) -> None:
        rate_limit_error: Optional[opik.exceptions.OpikCloudRequestsRateLimited] = None

        for processor in self._processors:
            try:
                processor.process(message)
            except opik.exceptions.OpikCloudRequestsRateLimited as ex:
                rate_limit_error = ex
            except Exception as ex:
                LOGGER.error(
                    "Unexpected error while processing message: %s with message processor: %s",
                    ex,
                    type(processor),
                    exc_info=True,
                )

        # Rate limit error is a special case that is handled by the caller.
        if rate_limit_error is not None:
            raise rate_limit_error

    def get_processor_by_type(self, processor_type: Type[T]) -> Optional[T]:
        """
        Retrieves a processor from the available processors that matches the specified type.

        This method iterates through the list of processors and checks if any of them is
        an instance of the given class type. If a match is found, it returns the processor.

        Args:
            processor_type: Concrete processor class to search for.

        Returns:
            The processor matching the specified type if found, else None.
        """
        for processor in self._processors:
            if isinstance(processor, processor_type):
                return processor
        return None

    def add_first(self, processor: BaseMessageProcessor) -> None:
        """
        Inserts a processor at the first position in the list of processors.

        This method allows prioritizing a given processor by placing it
        at the beginning of the internal processor list. As a result, the
        provided processor will be executed before others.

        Args:
            processor: The message processor to be
                added to the front of the processor list.
        """
        self._processors.insert(0, processor)



# This file was auto-generated by Fern from our API Definition.

import typing

from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from ..types.agent_blueprint_public import AgentBlueprintPublic
from ..types.agent_blueprint_write import AgentBlueprintWrite
from ..types.agent_config_env import AgentConfigEnv
from ..types.blueprint_page_history import BlueprintPageHistory
from .raw_client import AsyncRawAgentConfigsClient, RawAgentConfigsClient

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class AgentConfigsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawAgentConfigsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawAgentConfigsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawAgentConfigsClient
        """
        return self._raw_client

    def create_agent_config(
        self,
        *,
        blueprint: AgentBlueprintWrite,
        project_id: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Creates a new optimizer config with initial blueprint. Fails if the project already has a config.

        Parameters
        ----------
        blueprint : AgentBlueprintWrite

        project_id : typing.Optional[str]
            Project ID. Either project_id or project_name must be provided

        project_name : typing.Optional[str]
            Project name. Either project_id or project_name must be provided

        id : typing.Optional[str]
            Agent config ID. Generated automatically if not provided

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        from Opik import AgentBlueprintWrite
        from Opik import AgentConfigValueWrite
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.agent_configs.create_agent_config(blueprint=AgentBlueprintWrite(type="blueprint", values=[AgentConfigValueWrite(key='key', type="string", )], ), )
        """
        _response = self._raw_client.create_agent_config(
            blueprint=blueprint,
            project_id=project_id,
            project_name=project_name,
            id=id,
            request_options=request_options,
        )
        return _response.data

    def update_agent_config(
        self,
        *,
        blueprint: AgentBlueprintWrite,
        project_id: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Adds a new blueprint to an existing optimizer config. Fails if the project has no config yet.

        Parameters
        ----------
        blueprint : AgentBlueprintWrite

        project_id : typing.Optional[str]
            Project ID. Either project_id or project_name must be provided

        project_name : typing.Optional[str]
            Project name. Either project_id or project_name must be provided

        id : typing.Optional[str]
            Agent config ID. Generated automatically if not provided

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        from Opik import AgentBlueprintWrite
        from Opik import AgentConfigValueWrite
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.agent_configs.update_agent_config(blueprint=AgentBlueprintWrite(type="blueprint", values=[AgentConfigValueWrite(key='key', type="string", )], ), )
        """
        _response = self._raw_client.update_agent_config(
            blueprint=blueprint,
            project_id=project_id,
            project_name=project_name,
            id=id,
            request_options=request_options,
        )
        return _response.data

    def create_or_update_envs(
        self,
        *,
        project_id: str,
        envs: typing.Sequence[AgentConfigEnv],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Creates or updates environment-to-blueprint mappings

        Parameters
        ----------
        project_id : str

        envs : typing.Sequence[AgentConfigEnv]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        from Opik import AgentConfigEnv
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.agent_configs.create_or_update_envs(project_id='project_id', envs=[AgentConfigEnv(env_name='env_name', blueprint_id='blueprint_id', )], )
        """
        _response = self._raw_client.create_or_update_envs(
            project_id=project_id, envs=envs, request_options=request_options
        )
        return _response.data

    def get_blueprint_by_env(
        self,
        env_name: str,
        project_id: str,
        *,
        mask_id: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AgentBlueprintPublic:
        """
        Retrieves the blueprint associated with a specific environment

        Parameters
        ----------
        env_name : str

        project_id : str

        mask_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AgentBlueprintPublic
            Blueprint retrieved

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.agent_configs.get_blueprint_by_env(env_name='env_name', project_id='project_id', )
        """
        _response = self._raw_client.get_blueprint_by_env(
            env_name, project_id, mask_id=mask_id, request_options=request_options
        )
        return _response.data

    def set_env_by_blueprint_name(
        self,
        env_name: str,
        project_id: str,
        *,
        blueprint_name: str,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Sets an environment to point to a blueprint identified by name

        Parameters
        ----------
        env_name : str

        project_id : str

        blueprint_name : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.agent_configs.set_env_by_blueprint_name(env_name='env_name', project_id='project_id', blueprint_name='blueprint_name', )
        """
        _response = self._raw_client.set_env_by_blueprint_name(
            env_name, project_id, blueprint_name=blueprint_name, request_options=request_options
        )
        return _response.data

    def delete_env(
        self, env_name: str, project_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Soft-deletes an environment by setting its ended_at timestamp

        Parameters
        ----------
        env_name : str

        project_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.agent_configs.delete_env(env_name='env_name', project_id='project_id', )
        """
        _response = self._raw_client.delete_env(env_name, project_id, request_options=request_options)
        return _response.data

    def get_blueprint_by_id(
        self,
        blueprint_id: str,
        *,
        mask_id: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AgentBlueprintPublic:
        """
        Retrieves a specific blueprint by its ID

        Parameters
        ----------
        blueprint_id : str

        mask_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AgentBlueprintPublic
            Blueprint retrieved

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.agent_configs.get_blueprint_by_id(blueprint_id='blueprint_id', )
        """
        _response = self._raw_client.get_blueprint_by_id(blueprint_id, mask_id=mask_id, request_options=request_options)
        return _response.data

    def get_blueprint_by_name(
        self,
        project_id: str,
        name: str,
        *,
        mask_id: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AgentBlueprintPublic:
        """
        Retrieves a specific blueprint by its name within a project

        Parameters
        ----------
        project_id : str

        name : str

        mask_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AgentBlueprintPublic
            Blueprint retrieved

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.agent_configs.get_blueprint_by_name(project_id='project_id', name='name', )
        """
        _response = self._raw_client.get_blueprint_by_name(
            project_id, name, mask_id=mask_id, request_options=request_options
        )
        return _response.data

    def get_blueprint_history(
        self,
        project_id: str,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> BlueprintPageHistory:
        """
        Retrieves paginated blueprint history for a project

        Parameters
        ----------
        project_id : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        BlueprintPageHistory
            History retrieved

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.agent_configs.get_blueprint_history(project_id='project_id', )
        """
        _response = self._raw_client.get_blueprint_history(
            project_id, page=page, size=size, request_options=request_options
        )
        return _response.data

    def get_delta_by_id(
        self, blueprint_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AgentBlueprintPublic:
        """
        Retrieves only the changes (delta) introduced in a specific blueprint

        Parameters
        ----------
        blueprint_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AgentBlueprintPublic
            Delta retrieved

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.agent_configs.get_delta_by_id(blueprint_id='blueprint_id', )
        """
        _response = self._raw_client.get_delta_by_id(blueprint_id, request_options=request_options)
        return _response.data

    def get_latest_blueprint(
        self,
        project_id: str,
        *,
        mask_id: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AgentBlueprintPublic:
        """
        Retrieves the latest blueprint for a project

        Parameters
        ----------
        project_id : str

        mask_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AgentBlueprintPublic
            Blueprint retrieved

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.agent_configs.get_latest_blueprint(project_id='project_id', )
        """
        _response = self._raw_client.get_latest_blueprint(project_id, mask_id=mask_id, request_options=request_options)
        return _response.data


class AsyncAgentConfigsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawAgentConfigsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawAgentConfigsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawAgentConfigsClient
        """
        return self._raw_client

    async def create_agent_config(
        self,
        *,
        blueprint: AgentBlueprintWrite,
        project_id: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Creates a new optimizer config with initial blueprint. Fails if the project already has a config.

        Parameters
        ----------
        blueprint : AgentBlueprintWrite

        project_id : typing.Optional[str]
            Project ID. Either project_id or project_name must be provided

        project_name : typing.Optional[str]
            Project name. Either project_id or project_name must be provided

        id : typing.Optional[str]
            Agent config ID. Generated automatically if not provided

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        from Opik import AgentBlueprintWrite
        from Opik import AgentConfigValueWrite
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.agent_configs.create_agent_config(blueprint=AgentBlueprintWrite(type="blueprint", values=[AgentConfigValueWrite(key='key', type="string", )], ), )
        asyncio.run(main())
        """
        _response = await self._raw_client.create_agent_config(
            blueprint=blueprint,
            project_id=project_id,
            project_name=project_name,
            id=id,
            request_options=request_options,
        )
        return _response.data

    async def update_agent_config(
        self,
        *,
        blueprint: AgentBlueprintWrite,
        project_id: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Adds a new blueprint to an existing optimizer config. Fails if the project has no config yet.

        Parameters
        ----------
        blueprint : AgentBlueprintWrite

        project_id : typing.Optional[str]
            Project ID. Either project_id or project_name must be provided

        project_name : typing.Optional[str]
            Project name. Either project_id or project_name must be provided

        id : typing.Optional[str]
            Agent config ID. Generated automatically if not provided

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        from Opik import AgentBlueprintWrite
        from Opik import AgentConfigValueWrite
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.agent_configs.update_agent_config(blueprint=AgentBlueprintWrite(type="blueprint", values=[AgentConfigValueWrite(key='key', type="string", )], ), )
        asyncio.run(main())
        """
        _response = await self._raw_client.update_agent_config(
            blueprint=blueprint,
            project_id=project_id,
            project_name=project_name,
            id=id,
            request_options=request_options,
        )
        return _response.data

    async def create_or_update_envs(
        self,
        *,
        project_id: str,
        envs: typing.Sequence[AgentConfigEnv],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Creates or updates environment-to-blueprint mappings

        Parameters
        ----------
        project_id : str

        envs : typing.Sequence[AgentConfigEnv]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        from Opik import AgentConfigEnv
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.agent_configs.create_or_update_envs(project_id='project_id', envs=[AgentConfigEnv(env_name='env_name', blueprint_id='blueprint_id', )], )
        asyncio.run(main())
        """
        _response = await self._raw_client.create_or_update_envs(
            project_id=project_id, envs=envs, request_options=request_options
        )
        return _response.data

    async def get_blueprint_by_env(
        self,
        env_name: str,
        project_id: str,
        *,
        mask_id: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AgentBlueprintPublic:
        """
        Retrieves the blueprint associated with a specific environment

        Parameters
        ----------
        env_name : str

        project_id : str

        mask_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AgentBlueprintPublic
            Blueprint retrieved

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.agent_configs.get_blueprint_by_env(env_name='env_name', project_id='project_id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_blueprint_by_env(
            env_name, project_id, mask_id=mask_id, request_options=request_options
        )
        return _response.data

    async def set_env_by_blueprint_name(
        self,
        env_name: str,
        project_id: str,
        *,
        blueprint_name: str,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Sets an environment to point to a blueprint identified by name

        Parameters
        ----------
        env_name : str

        project_id : str

        blueprint_name : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.agent_configs.set_env_by_blueprint_name(env_name='env_name', project_id='project_id', blueprint_name='blueprint_name', )
        asyncio.run(main())
        """
        _response = await self._raw_client.set_env_by_blueprint_name(
            env_name, project_id, blueprint_name=blueprint_name, request_options=request_options
        )
        return _response.data

    async def delete_env(
        self, env_name: str, project_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Soft-deletes an environment by setting its ended_at timestamp

        Parameters
        ----------
        env_name : str

        project_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.agent_configs.delete_env(env_name='env_name', project_id='project_id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.delete_env(env_name, project_id, request_options=request_options)
        return _response.data

    async def get_blueprint_by_id(
        self,
        blueprint_id: str,
        *,
        mask_id: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AgentBlueprintPublic:
        """
        Retrieves a specific blueprint by its ID

        Parameters
        ----------
        blueprint_id : str

        mask_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AgentBlueprintPublic
            Blueprint retrieved

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.agent_configs.get_blueprint_by_id(blueprint_id='blueprint_id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_blueprint_by_id(
            blueprint_id, mask_id=mask_id, request_options=request_options
        )
        return _response.data

    async def get_blueprint_by_name(
        self,
        project_id: str,
        name: str,
        *,
        mask_id: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AgentBlueprintPublic:
        """
        Retrieves a specific blueprint by its name within a project

        Parameters
        ----------
        project_id : str

        name : str

        mask_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AgentBlueprintPublic
            Blueprint retrieved

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.agent_configs.get_blueprint_by_name(project_id='project_id', name='name', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_blueprint_by_name(
            project_id, name, mask_id=mask_id, request_options=request_options
        )
        return _response.data

    async def get_blueprint_history(
        self,
        project_id: str,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> BlueprintPageHistory:
        """
        Retrieves paginated blueprint history for a project

        Parameters
        ----------
        project_id : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        BlueprintPageHistory
            History retrieved

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.agent_configs.get_blueprint_history(project_id='project_id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_blueprint_history(
            project_id, page=page, size=size, request_options=request_options
        )
        return _response.data

    async def get_delta_by_id(
        self, blueprint_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AgentBlueprintPublic:
        """
        Retrieves only the changes (delta) introduced in a specific blueprint

        Parameters
        ----------
        blueprint_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AgentBlueprintPublic
            Delta retrieved

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.agent_configs.get_delta_by_id(blueprint_id='blueprint_id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_delta_by_id(blueprint_id, request_options=request_options)
        return _response.data

    async def get_latest_blueprint(
        self,
        project_id: str,
        *,
        mask_id: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AgentBlueprintPublic:
        """
        Retrieves the latest blueprint for a project

        Parameters
        ----------
        project_id : str

        mask_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AgentBlueprintPublic
            Blueprint retrieved

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.agent_configs.get_latest_blueprint(project_id='project_id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_latest_blueprint(
            project_id, mask_id=mask_id, request_options=request_options
        )
        return _response.data



from .tts_create_decorator import TTSCreateTrackDecorator
from .tts_streaming_response_decorator import TTSStreamingResponseCreateTrackDecorator

__all__ = [
    "TTSCreateTrackDecorator",
    "TTSStreamingResponseCreateTrackDecorator",
]



# This file was auto-generated by Fern from our API Definition.

import typing

ProjectMetricResponsePublicMetricType = typing.Union[
    typing.Literal[
        "FEEDBACK_SCORES",
        "TRACE_COUNT",
        "TOKEN_USAGE",
        "DURATION",
        "COST",
        "GUARDRAILS_FAILED_COUNT",
        "THREAD_COUNT",
        "THREAD_DURATION",
        "THREAD_FEEDBACK_SCORES",
        "SPAN_FEEDBACK_SCORES",
        "SPAN_COUNT",
        "SPAN_DURATION",
        "SPAN_TOKEN_USAGE",
    ],
    typing.Any,
]



from .base import BaseSuiteEvaluator
from .llm_judge import LLMJudge, LLMJudgeConfig, DEFAULT_MODEL_NAME

__all__ = [
    "BaseSuiteEvaluator",
    "LLMJudge",
    "LLMJudgeConfig",
    "DEFAULT_MODEL_NAME",
]
